local _G = _G
local L = PetSelector.L
local GetSpellInfo = GetSpellInfo
local UnitGUID = UnitGUID
local UnitName = UnitName
local UnitClass = UnitClass
local grey = "|cff778899"
local white = "|cffffffff"

--find an available pet that matches the supplied family
local function findPet(family, pets)

	local v, pet
	
	for _, v in ipairs(pets) do

		if v.family == family then

			if (v.slot < 5) or PetSelector.db.profile.stabled then
						
				pet = v
				
			end
				
		end
			
	end
		
	return pet
	
end

--scan the potentially available buffs and select the most
--appropriate pet
local function getPetBuff(family)

	local pbuff, buff
	
	for _, pbuff in ipairs(PetSelector.PetInfo) do
	
		if pbuff[1] == family then
		
			buff = pbuff[2]
			break
			
		end
		
	end
	
	return buff
	
end

local function setBuff(index, pclass, player, petfamily)

	local sicon, sid, buff, pc
	
	if PetSelector.debug then
	
		--print("buffi: " .. (index or "nill") .. " bc: " .. (pclass or "nil") .. " bp: " .. (player or "nil") .. " bpet: " .. (pet or "nil"))
		
	end
	
	if pclass then
	
		if pclass == "HUNTER" then
		
			sid = getPetBuff(petfamily)
			
		else
		
			buff = PetSelector.PlayerBuffs[index]
		
			for _, pc in ipairs(buff) do
			
				if pc[1] == pclass then
				
					sid = pc[4]
					break
					
				end
				
			end
				
		end
			
	else
	
		sid = PetSelector.PlayerBuffs[index][1][4]
		
	end
	
	if not sid then return end
	
	if (index == 1) and (PetSelector.faction == "Horde") and (pclass == "SHAMAN") then
		
		sid = PetSelector.PlayerBuffs[index][1][5]
		sicon = select(3, GetSpellInfo(sid))
		
	else
	
		sicon = select(3, GetSpellInfo(sid))
		
	end
		
	_G["PetSelectorBuffIcon" .. index]:SetTexture(sicon)
	_G["PetSelectorBuff" .. index].spellid = sid
	_G["PetSelectorBuff" .. index].player = player

	if player then
	
		_G["PetSelectorBuffProvider" .. index]:SetText(white .. player)
		
	else
	
		_G["PetSelectorBuffProvider" .. index]:SetText(grey .. "<" .. L["not detected"] .. ">")
		
	end
		
end

function PetSelector:SelectPet()

	local buff, petfamily
	local isBM = (select(5, self.GT:GetUnitTalentSpec("player")) == "HunterBeastMastery")
	local start = isBM and 1 or 4
	local pet, rnd, v, t, index
	local pets = self:GetPets()
	
	self.random = nil
	
	if self.inparty then
	
		for buff = start, #self.Buffs do
	
			if self.Buffs[buff] == false then
		
				for _, petfamily in pairs(self.PetBuffs[buff]) do
			
					if not isBM then
				
						if not self:IsExotic(petfamily) then
					
							pet = findPet(petfamily, pets)
							index = buff
							if pet then break end
						
						end
					
					else

						pet = findPet(petfamily, pets)
						index = buff
						if pet then break end
					
					end
				
				end
			
			end
		
			if pet then break end

		end
	
	end
	
	if (not pet) and (#pets > 0) then
	
		--no suitable pet or all buffs catered for
		--pick a random pet from those not in the stable
		t = 1
		
		while true do
		
			rnd = math.random(25)

			if pets[rnd] then

				if pets[rnd].slot < 6 then
					
					if self.selectedpet then
					
						if self.selectedpet.name ~= pets[rnd].name then
						
							pet = pets[rnd] break
							
						end
						
					else
					
						pet = pets[rnd]
						break
						
					end
					
				end
				
			end
		
			--prevent an infinite loop in the unlikely event there are no pets outside the stable
			t = t + 1
			if t > 50 then break end
			
		end
		
		self.random = true
		
	elseif pet then
	
		setBuff(index, "HUNTER", UnitName("player"), pet.family)
		
	end
	
	self.selectedpet = pet
	self:SetPetMacro()
	return pet

end

--build a table representing potentially available buffs
local function getHunterPet(pname, players, pre)

	local prefix = pre .. "pet"
	local index, p, petname
	
	for p = 1, players do
	
		if UnitName(pre .. p) == pname then
		
			pet = UnitCreatureType(prefix .. p)
			break
			
		end
		
	end
	
	if PetSelector.debug then
	
		print("hunter: " .. (pname or "nil") .. " pet: " .. (pet or "nil"))
		
	end

	return pet
	
end

local function checkHunterPet(index, pet)

	local family, found
	
	if not pet then return end
	
	for _, family in ipairs(PetSelector.PetBuffs[index]) do
	
		if family == pet then
		
			found = true
			break
			
		end
		
	end
	
	return found
	
end

local function CheckTree(...)

	local t1 = select(2, ...)
	local t2 = select(3, ...)
	local t3 = select(4, ...)
	local t = 1
	
	if t2 > t1 and t2 > t3 then t = 2 end
	if t3 > t1 and t3 > t2 then t = 3 end
	
	return t
	
end

local function IsExcluded(player, buff)

	local ex, isex
	
	for _, ex in ipairs(PetSelector.exclude) do
	
		if (ex.player == player) and (ex.id == buff) then
		
			isex = true
			break
			
		end
		
	end
	
	return isex
	
end

function PetSelector:CheckBuffs()

	local guid, class, pclass, pname
	local buff, n, tree, talent, player
	local players = GetNumRaidMembers()
	local prefix = "raid"
	local sicon, pet, hbuff, s
	
	hbuff = {}
	
	if players == 0 then
	
		players = GetNumPartyMembers()
		prefix = "party"
		
	end
	
	--reinitialise buff table
	for buff = 1, 12 do
	
		self.Buffs[buff] = false
		setBuff(buff)
		hbuff[buff] = false
		
	end

	--no point checking the buffs if we're not in a group
	if (not self.inparty) or (players == 0) then return end
		
	--rebuild buff table
	for player = 1, players do
	
		if (player == 1) and self.debug then
		
			--print("Building buff table")
			
		end
		
		guid = UnitGUID(prefix .. player)
		pname = UnitName(prefix .. player)
		class = select(2, UnitClass(prefix .. player))
		
		if class == "HUNTER" then
		
			pet = getHunterPet(pname, players, prefix)
			
		end
		
		for n, buff in pairs(self.PlayerBuffs) do
			
			if (not self.Buffs[n]) or hbuff[n] then

				for _, pclass in ipairs(buff) do
			
					if not IsExcluded(pname, n) then
					
						if pclass[1] == class then
				
							if class == "HUNTER" then
			
								if checkHunterPet(n, pet) then
				
									self.Buffs[n] = true
									hbbuff[n] = true
									setBuff(n, class, pname, pet)
					
									if self.debug then
						
										--print(pname .. ": " .. pclass[4])
							
									end
							
								end
				
								break
							
							elseif pclass[2] == nil then	--no tree
					
								self.Buffs[n] = true
								hbuff[n] = false
								setBuff(n, class, pname)
						
								if self.debug then
						
									--print(pname .. ": " .. pclass[4])
							
								end
						
								break
						
							elseif pclass[3] == nil then	--tree, but no talent
					
								if CheckTree(self.GT:GetGUIDTalentSpec(guid)) == pclass[2] then
						
									self.Buffs[n] = true
									hbuff[n] = false
									setBuff(n, class, pname)
						
									if self.debug then
						
										--print(pname .. ": " .. pclass[4])
							
									end
							
									break
							
								end
						
							else	--tree and talent
					
								s = GetSpellInfo(pclass[3])

								if self.GT:GUIDHasTalent(guid, s) then
						
									self.Buffs[n] = true
									hbuff[n] = false
									setBuff(n, class, pname)
							
									if self.debug then
						
										--print(pname .. ": " .. pclass[4])
							
									end
						
									break
							
								end
						
							end
						
						end
					
					end
				
				end
			
			end
	
		end
	
	end

end

--retrieve basic pet info
local function getInfo(pet)

	local text, family
	local petfamily
	
	for _, family in ipairs(PetSelector.PetInfo) do

		if family[1] == pet.family then
		
			petfamily = family
			break
			
		end
		
	end
	
	return petfamily
	
end

local function EnumerateTooltipLines_helper(...)

    for i = 1, select("#", ...) do
	
        local region = select(i, ...)
		
        if region and region:GetObjectType() == "FontString" then
		
            local text = region:GetText()
			DEFAULT_CHAT_FRAME:AddMessage(i .. ": " .. (text or "nil"), 0, 1, 0)
			
        end
		
    end
	
end

local function EnumerateTooltipLines(tooltip)

    EnumerateTooltipLines_helper(tooltip:GetRegions())
	
end

--main frame
function PetSelector:CreatePetFrame()

	local frame = CreateFrame("Frame", "PetSelectorMainFrame", UIParent)
	self.button = CreateFrame("Button", nil, frame)
	self.pushed = self.button:CreateTexture()
	self.tooltip = self.tooltip or CreateFrame("GameTooltip", "PetSelector_Tooltip", nil, "GameTooltipTemplate")
	
	local function buttonOnEnter(this)
	
		local pet = PetSelector.selectedpet
		
		if not pet then return end
		
		local info = getInfo(pet)
		local regions, spellname, casttime, cd, desc
		local stabled, rand = ""
		
		--get the spell info
		self.tooltip:SetOwner(UIParent, "ANCHOR_NONE")
		self.tooltip:ClearLines()
		self.tooltip:SetSpellByID(info[2])
		regions = {self.tooltip:GetRegions()}
        spellname = regions[10]:GetText()
		
		if(regions[15]:GetText()) then
		
			casttime = regions[14]:GetText()
			cd = regions[15]:GetText()
			desc = regions[16]: GetText()
			
		else
			
			casttime = regions[12]:GetText()
			cd = regions[13]:GetText()
			desc = regions[14]:GetText()
			
		end
		
		if self.debug then
		
			--EnumerateTooltipLines(self.tooltip)
			
		end
		
		self.tooltip:Hide()
		
		--is the pet in the stable?
		if pet.slot > 5 then
		
			stabled = "   |cffcc3333" .. L["stabled"] .. "|r"
			
		end
		
		--setup the tooltip
		GameTooltip:SetOwner(this, "ANCHOR_TOP")
		GameTooltip:ClearLines()
		GameTooltip:SetText("Pet Selector")
		GameTooltip:AppendText("\r\r|cff00ff00" .. pet.name .. "|r" .. stabled)
		
		--is this a random selection pet?
		if self.random then
		
			GameTooltip:AddDoubleLine(pet.family, L["Random selection"], 1, 1, 1, 0.25, 0.3, 0.8)
			
		else
		
			GameTooltip:AddLine(pet.family, 1, 1, 1)
			
		end
		
		GameTooltip:AddLine(" ")
		GameTooltip:AddDoubleLine(spellname, casttime, nil, nil, nil, 1, 1, 1)
		GameTooltip:AddLine(cd, 0.1, 0.6, 1)
		GameTooltip:AddLine(desc, 1, 1, 1, true)
		GameTooltip:Show()
		
	end
	
	local function buttonOnLeave(this)
	
		GameTooltip:Hide()
		
	end
	
	local function frameDragStop()
		
		frame:StopMovingOrSizing()
		self.db.profile.x = frame:GetLeft()
		self.db.profile.y = frame:GetBottom()

	end
	
	frame:Hide()
	frame:SetMovable(true)
	frame:EnableMouse(true)
	frame:EnableKeyboard(true)
	frame:SetClampedToScreen(true)
	frame:SetFrameStrata("MEDIUM")
	frame:SetHeight(32)
	frame:SetWidth(32)
	frame:RegisterForDrag("LeftButton")
	frame:SetScript("OnDragStart", function() if not self.db.profile.locked then frame:StartMoving() end end)
	frame:SetScript("OnDragStop", frameDragStop)
	frame:SetScale(self.db.profile.scale)
	frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", self.db.profile.x, self.db.profile.y)
	
	self.pushed:SetAllPoints(self.button)
	self.pushed:SetAlpha(0.5)
	
	local function onClick(this, button, updown)
	
		if button == "LeftButton" then
		
			GameTooltip:Hide()
			self:SetPet()
			buttonOnEnter(this)
			
		else
			
			InterfaceOptionsFrame_OpenToCategory(self.optionschecks)
			
		end
		
	end
	
	self.button:SetWidth(32)
	self.button:SetHeight(32)
	self.button:SetAllPoints(frame)
	self.button:SetHighlightTexture("Interface\\BUTTONS\\ButtonHilight-Square", "ADD")
	self.button:SetPushedTexture(self.pushed)
	self.button:RegisterForDrag("LeftButton")
	self.button:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	self.button:SetScript("OnEnter", buttonOnEnter)
	self.button:SetScript("OnLeave", buttonOnLeave)	
	self.button:SetScript("OnDragStart", function() if not self.db.profile.locked then frame:StartMoving() end end)
	self.button:SetScript("OnDragStop", frameDragStop)
	self.button:SetScript("OnClick", onClick)
	
end

local function createLine(lnum, parent)

	local frame = CreateFrame("Button", "PetSelectorBuff" .. lnum, parent)
	local icon = frame:CreateTexture("PetSelectorBuffIcon" .. lnum)
	local provider = frame:CreateFontString("PetSelectorBuffProvider" .. lnum, "ARTWORK", "GameFontNormal")
	
	frame:SetWidth(parent:GetWidth())
	frame:SetHeight(16)
	frame:SetID(lnum)
	frame:RegisterForClicks("AnyUp")

	local function onEnter(this)
	
		GameTooltip:SetOwner(this, "ANCHOR_LEFT")
		GameTooltip:ClearLines()
		GameTooltip:SetSpellByID(this.spellid)
		
	end
	
	local function onLeave(this)
	
		GameTooltip:Hide()
		
	end
	
	local function onClick(this, button)
	
		local id, provider, dd
		
		if button == "RightButton" then

			--do drop down menu
			PetSelectorMenu:SetParent(this)
			ToggleDropDownMenu(1, nil, PetSelectorMenu, "cursor", 3, -3)

		end
		
	end
	
	frame:SetScript("OnEnter", onEnter)
	frame:SetScript("OnLeave", onLeave)
	frame:SetScript("OnClick", onClick)

	icon:SetWidth(16)
	icon:SetHeight(16)
	icon:SetPoint("LEFT", frame, "LEFT")
	
	provider:SetWidth(frame:GetWidth() - 8 - 16 - 5)
	provider:SetHeight(16)
	provider:SetPoint("LEFT", icon, "RIGHT", 0, 0)
	
	return frame

end

function PetSelector:DisplayFrames()

	local db = self.db.profile
	local displaymain = ((db.showinparty and self.inparty) or (not db.showinparty))
	local displaybuffs = db.showbuffs and displaymain
	
	if displaymain then
	
		PetSelectorMainFrame:Show()
		
	else
	
		PetSelectorMainFrame:Hide()
		
	end
	
	if ((displaybuffs == true ) and (self.inparty or self.debug)) then
	
		PetSelectorBuffFrame:Show()
		
	else
	
		PetSelectorBuffFrame:Hide()
		
	end
	
end

function PetSelector:CreateDropDown()

	local menu = CreateFrame("Frame", "PetSelectorMenu")

	menu.displayMode = "MENU"
	menu.info = {}
	menu.HideMenu = 
		function()
			if UIDROPDOWNMENU_OPEN_MENU == menu then
				CloseDropDownMenus()
			end
		end

	menu.initialize = 
		function(this, level)
		
			if not level then return end
			
			local info = this.info
			local id = this:GetParent():GetID()
			local ex

			wipe(info)
			
			if level == 1 then
			
				info.isTitle = 1
				info.text = "Pet Selector"
				info.notCheckable = 1
				UIDropDownMenu_AddButton(info, level)

				info.keepShownOnClick = 1
				info.disabled = nil
				info.isTitle = nil

				info.text = L["Ignore player"]
				info.hasArrow = nil
				info.func = function() table.insert(PetSelector.exclude, {id = id, player = this.player}); PetSelector:CheckBuffs() end
				info.disabled = not this.player
				UIDropDownMenu_AddButton(info, level)

				info.text = L["Unignore player"]
				info.value = "unignore"
				info.disabled = true
				info.hasArrow = nil
				
				for _, ex in ipairs(self.exclude) do
				
					if ex.id == id then
				
						info.disabled = nil
						info.hasArrow = 1
						break;
						
					end
					
				end
				
				UIDropDownMenu_AddButton(info, level)

				-- Close menu item
				info.hasArrow = nil
				info.value = nil
				info.text = _G.CLOSE
				info.func = this.HideMenu
				info.disabled = nil
				UIDropDownMenu_AddButton(info, level)

			elseif level == 2 then
			
				if (UIDROPDOWNMENU_MENU_VALUE == "unignore") then
				
					for _, ex in ipairs(self.exclude) do
					
						info.text = ex.player
						info.notCheckable = 1
						info.func = 
							function(this)
								local i, e
								for i, e in ipairs(PetSelector.exclude) do
									if e.player == this.info.text then
										PetSelector.exclude[i] = nil
										PetSelector:CheckBuffs()
										break
									end
								end
							end
						UIDropDownMenu_AddButton(info, level)
						
					end
					
				end
					
			end
			
		end

end

function PetSelector:CreateBuffFrame()

	local frame = CreateFrame("Frame", "PetSelectorBuffFrame", UIParent)
	local closeButton = CreateFrame("Button", nil, frame, "UIPanelCloseButton")
	local title = frame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	local line, fname, tmp
	local ypos = -35
	
	local function frameDragStop()
		
		frame:StopMovingOrSizing()
		self.db.profile.bx = frame:GetLeft()
		self.db.profile.by = frame:GetBottom()

	end
	
	frame:Hide()
	frame:SetMovable(true)
	frame:EnableMouse(true)
	frame:EnableKeyboard(true)
	frame:SetClampedToScreen(true)
	frame:SetFrameStrata("MEDIUM")
	frame:SetHeight(280)
	frame:SetWidth(150)
	frame:RegisterForDrag("LeftButton")
	frame:SetScript("OnDragStart", function() if not self.db.profile.locked then frame:StartMoving() end end)
	frame:SetScript("OnDragStop", frameDragStop)
	frame:SetScale(self.db.profile.scale)
	frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", self.db.profile.bx, self.db.profile.by)
	frame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 16, edgeSize = 16, insets = {left = 4, right = 4, top = 4, bottom = 4}})
	frame:SetBackdropColor(0, 0, 0, 1)

	local function closeClick()
	
		frame:Hide()
		self.db.profile.showbuffs = false
		
		if (self.optionschecks:IsVisible()) then
		
			InterfaceOptionsFrame:Hide()
			InterfaceOptionsFrame_OpenToCategory(self.optionschecks)
			
		end
		
	end
		
	closeButton:SetPoint("TOPLEFT", frame, "TOPRIGHT", -32, -2)
	closeButton:SetScript("OnClick", closeClick)
	
	title:SetPoint("CENTER", frame, "CENTER", -16, 125)
	title:SetText(L["Buff Providers"])
	
	for line = 1, 12 do
	
		tmp = createLine(line, frame)
		tmp:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, ypos)
		ypos = ypos - 20
		
	end
	
	self:CreateDropDown()

end